APP_NAME = 'Unmixer'
ORGANIZATION_NAME = 'DyeSoft'
PROJECT_README_URL = 'https://github.com/will2dye4/unmixer#usage'

ERROR_MESSAGE_TITLE = 'Error'
SUCCESS_MESSAGE_TITLE = 'Success'

OK_BUTTON_TEXT = 'OK'

OTHER_TRACK_NAME_DISABLED_TOOLTIP = 'Select a different pre-trained model to change this setting.'

FONT_WEIGHT_BOLD = 800
